package eu.mhsl.craftattack.spawn.appliance;

import eu.mhsl.craftattack.spawn.Main;
import eu.mhsl.craftattack.spawn.api.HttpServer;
import eu.mhsl.craftattack.spawn.config.Configuration;
import org.bukkit.Bukkit;
import org.bukkit.command.PluginCommand;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.event.HandlerList;
import org.bukkit.event.Listener;
import org.bukkit.plugin.java.JavaPlugin;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 * Any implementation of this class can be seen as a "sub-plugin" with its own event handlers and commands.
 * Appliances can be enabled or disabled independent of other appliances
 */
public abstract class Appliance {
    private String localConfigPath;
    private List<Listener> listeners;
    private List<ApplianceCommand<?>> commands;

    public Appliance() {
    }

    /**
     * Use this constructor to specify a config sub-path for use with the localConfig() method.
     *
     * @param localConfigPath sub path, if not found, the whole config will be used
     */
    public Appliance(String localConfigPath) {
        this.localConfigPath = localConfigPath;
    }

    /**
     * Provides a list of listeners for the appliance. All listeners will be automatically registered.
     *
     * @return List of listeners
     */
    @NotNull
    protected List<Listener> eventHandlers() {
        return new ArrayList<>();
    }

    /**
     * Provides a list of commands for the appliance. All commands will be automatically registered.
     *
     * @return List of commands
     */
    @NotNull
    protected List<ApplianceCommand<?>> commands() {
        return new ArrayList<>();
    }

    /**
     * Called on initialization to add all needed API Routes.
     * The routeBuilder can be used to get the correct Path prefixes
     *
     * @param apiBuilder holds data for needed route prefixes.
     */
    public void httpApi(HttpServer.ApiBuilder apiBuilder) {
    }

    /**
     * Provides a localized config section. Path can be set in appliance constructor.
     *
     * @return Section of configuration for your appliance
     */
    @NotNull
    public ConfigurationSection localConfig() {
        return Optional.ofNullable(Configuration.cfg.getConfigurationSection(localConfigPath)).orElse(Configuration.cfg);
    }

    public void onEnable() {
    }

    public void onDisable() {
    }

    public void initialize(@NotNull JavaPlugin plugin) {
        this.listeners = eventHandlers();
        this.commands = commands();

        listeners.forEach(listener -> Bukkit.getPluginManager().registerEvents(listener, plugin));
        commands.forEach(command -> setCommandExecutor(plugin, command.commandName, command));
    }

    public void destruct(@NotNull JavaPlugin plugin) {
        listeners.forEach(HandlerList::unregisterAll);
    }

    public <T extends Appliance> T queryAppliance(Class<T> clazz) {
        return Main.instance().getAppliance(clazz);
    }

    private void setCommandExecutor(JavaPlugin plugin, String name, ApplianceCommand<?> executor) {
        PluginCommand command = plugin.getCommand(name);
        if(command != null && executor != null) {
            command.setExecutor(executor);
            command.setTabCompleter(executor);
        } else {
            Main.logger().warning("Command " + name + " is not specified in plugin.yml!");
            throw new RuntimeException("All commands must be registered in plugin.yml. Missing command: " + name);
        }
    }

    public List<Listener> getListeners() {
        return listeners;
    }

    public List<ApplianceCommand<?>> getCommands() {
        return commands;
    }
}
