package eu.mhsl.craftattack.spawn.appliances.projectStart;

import eu.mhsl.craftattack.spawn.appliance.Appliance;
import eu.mhsl.craftattack.spawn.appliance.ApplianceCommand;
import eu.mhsl.craftattack.spawn.appliances.projectStart.command.ProjectStartCancelCommand;
import eu.mhsl.craftattack.spawn.appliances.projectStart.command.ProjectStartCommand;
import eu.mhsl.craftattack.spawn.appliances.projectStart.command.ProjectStartResetCommand;
import eu.mhsl.craftattack.spawn.appliances.projectStart.listener.NoAdvancementsListener;
import eu.mhsl.craftattack.spawn.appliances.projectStart.listener.PlayerInvincibleListener;
import eu.mhsl.craftattack.spawn.config.Configuration;
import eu.mhsl.craftattack.spawn.util.IteratorUtil;
import eu.mhsl.craftattack.spawn.util.entity.PlayerUtils;
import eu.mhsl.craftattack.spawn.util.text.ComponentUtil;
import eu.mhsl.craftattack.spawn.util.text.Countdown;
import eu.mhsl.craftattack.spawn.util.world.BlockCycle;
import net.kyori.adventure.sound.Sound;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import org.bukkit.*;
import org.bukkit.entity.Player;
import org.bukkit.event.Listener;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.Map;
import java.util.Objects;

import static java.util.Map.entry;
import static org.bukkit.Sound.MUSIC_DISC_PIGSTEP;

public class ProjectStart extends Appliance {
    private final int startMusicAt = 137;
    private final World startWorld = Bukkit.getWorld("world");
    private final Location glassLocation = new Location(startWorld, -224, 67, -368);
    private final List<Location> netherFireLocations = List.of(
        new Location(startWorld, -211, 68, -354),
        new Location(startWorld, -209, 68, -356),
        new Location(startWorld, -207, 68, -354),
        new Location(startWorld, -209, 68, -352)
    );

    private final Countdown countdown = new Countdown(
        localConfig().getInt("countdown"),
        this::format,
        this::announce,
        this::startProject
    );
    private final BlockCycle blockCycle = new BlockCycle(
        glassLocation,
        Material.ORANGE_STAINED_GLASS,
        List.of(
            Material.RED_STAINED_GLASS,
            Material.YELLOW_STAINED_GLASS,
            Material.GREEN_STAINED_GLASS,
            Material.BLUE_STAINED_GLASS
        )
    );

    private final Map<GameRule<Boolean>, Boolean> gameRulesAfterStart = Map.ofEntries(
        entry(GameRule.DO_DAYLIGHT_CYCLE, true),
        entry(GameRule.DO_INSOMNIA, true),
        entry(GameRule.ANNOUNCE_ADVANCEMENTS, true),
        entry(GameRule.DISABLE_RAIDS, false),
        entry(GameRule.DO_FIRE_TICK, true),
        entry(GameRule.DO_ENTITY_DROPS, true),
        entry(GameRule.DO_PATROL_SPAWNING, true),
        entry(GameRule.DO_TRADER_SPAWNING, true),
        entry(GameRule.DO_WEATHER_CYCLE, true),
        entry(GameRule.FALL_DAMAGE, true),
        entry(GameRule.FIRE_DAMAGE, true)
    );

    public ProjectStart() {
        super("countdown");

        this.countdown.addCustomAnnouncement(
            new Countdown.CustomAnnouncements(
                counter -> counter == startMusicAt,
                counter -> glassLocation
                    .getWorld()
                    .playSound(glassLocation, MUSIC_DISC_PIGSTEP, SoundCategory.RECORDS, 500f, 1f)
            )
        );
    }

    private Component format(Countdown.AnnouncementData data) {
        return Component.text()
            .append(ComponentUtil.createRainbowText("CraftAttack", 10))
            .append(Component.text(" startet in ", NamedTextColor.GOLD))
            .append(Component.text(data.count(), NamedTextColor.AQUA))
            .append(Component.text(" " + data.unit() + "!", NamedTextColor.GOLD))
            .build();
    }

    private void announce(Component message) {
        blockCycle.next();
        IteratorUtil.onlinePlayers(player -> player.sendMessage(message));
    }

    public void startCountdown() {
        if(!isEnabled()) return;
        this.countdown.start();
    }

    public void cancelCountdown() {
        this.countdown.cancel();
        this.restoreBeforeStart();
    }

    public void startProject() {
        setEnabled(false);

        IteratorUtil.worlds(World::getWorldBorder, worldBorder -> worldBorder.setSize(worldBorder.getMaxSize()));
        IteratorUtil.worlds(world -> IteratorUtil.setGameRules(gameRulesAfterStart, false));
        IteratorUtil.worlds(world -> world.setFullTime(0));

        netherFireLocations.forEach(location -> Objects.requireNonNull(startWorld).getBlockAt(location).setType(Material.FIRE));

        Bukkit.getOnlinePlayers().forEach(player -> {
            player.setFoodLevel(20);
            player.setHealth(20);
            player.getInventory().clear();
            player.setGameMode(GameMode.SURVIVAL);
            player.setExp(0);
            player.setLevel(0);

            player.playSound(Sound.sound(org.bukkit.Sound.ITEM_GOAT_HORN_SOUND_5, Sound.Source.MASTER, 500f, 1f));

            player.sendMessage(Component.text("Viel Spaß bei CraftAttack!", NamedTextColor.GREEN));

            player.setStatistic(Statistic.TIME_SINCE_REST, 0);
            PlayerUtils.resetStatistics(player);
        });

        Bukkit.getServer().advancementIterator().forEachRemaining(
            advancement -> Bukkit.getOnlinePlayers().forEach(
                player -> player.getAdvancementProgress(advancement).getAwardedCriteria().forEach(
                    criteria -> player.getAdvancementProgress(advancement).revokeCriteria(criteria)
                )
            )
        );

        blockCycle.reset();
    }

    public void restoreBeforeStart() {
        setEnabled(true);

        IteratorUtil.onlinePlayers(Player::stopAllSounds);

        IteratorUtil.worlds(World::getWorldBorder, worldBorder -> {
            worldBorder.setSize(localConfig().getLong("worldborder-before"));
            worldBorder.setWarningDistance(0);
            worldBorder.setDamageAmount(0);
        });

        IteratorUtil.worlds(world -> world, world -> IteratorUtil.setGameRules(gameRulesAfterStart, true));


        blockCycle.reset();
    }

    public boolean isEnabled() {
        return localConfig().getBoolean("enabled");
    }

    public void setEnabled(boolean enabled) {
        localConfig().set("enabled", enabled);
        Configuration.saveChanges();
    }

    public Countdown getCountdown() {
        return countdown;
    }

    @Override
    @NotNull
    protected List<Listener> eventHandlers() {
        return List.of(
            new PlayerInvincibleListener(),
            new NoAdvancementsListener()
        );
    }

    @Override
    @NotNull
    protected List<ApplianceCommand<?>> commands() {
        return List.of(
            new ProjectStartCommand(),
            new ProjectStartCancelCommand(),
            new ProjectStartResetCommand()
        );
    }
}
